"""Extract info from a path following SSB's dataset naming convention."""

from __future__ import annotations

import logging
import pathlib
import re
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Final
from typing import Literal

import arrow

from datadoc.enums import DataSetState
from datadoc.enums import SupportedLanguages

if TYPE_CHECKING:
    import datetime
    import os
    from datetime import date

logger = logging.getLogger(__name__)


@dataclass
class DateFormat(ABC):
    """A super class for date formats."""

    name: str
    regex_pattern: str
    arrow_pattern: str
    timeframe: Literal["year", "month", "day", "week"]

    @abstractmethod
    def get_floor(self, period_string: str) -> date | None:
        """Return first date of timeframe period."""

    @abstractmethod
    def get_ceil(self, period_string: str) -> date | None:
        """Return last date of timeframe period."""


@dataclass
class IsoDateFormat(DateFormat):
    """A subclass of Dateformat with relevant patterns for ISO dates."""

    def get_floor(self, period_string: str) -> date | None:
        """Method.

        >>> ISO_YEAR_MONTH.get_floor("1980-08")
        datetime.date(1980, 8, 1)

        >>> ISO_YEAR.get_floor("2021")
        datetime.date(2021, 1, 1)

        >>> SSB_BIMESTER.get_floor("2003B4")
        datetime.date(2003, 7, 1)

        """
        return arrow.get(period_string, self.arrow_pattern).floor(self.timeframe).date()

    def get_ceil(self, period_string: str) -> date | None:
        """Method.

        >>> ISO_YEAR.get_ceil("1921")
        datetime.date(1921, 12, 31)

        >>> ISO_YEAR_MONTH.get_ceil("2021-05")
        datetime.date(2021, 5, 31)

        >>> SSB_HALF_YEAR.get_ceil("2024H1")
        datetime.date(2024, 6, 30)

        """
        return arrow.get(period_string, self.arrow_pattern).ceil(self.timeframe).date()


ISO_YEAR = IsoDateFormat(
    name="ISO_YEAR",
    regex_pattern=r"^\d{4}$",
    arrow_pattern="YYYY",
    timeframe="year",
)
ISO_YEAR_MONTH = IsoDateFormat(
    name="ISO_YEAR_MONTH",
    regex_pattern=r"^\d{4}\-\d{2}$",
    arrow_pattern="YYYY-MM",
    timeframe="month",
)
ISO_YEAR_MONTH_DAY = IsoDateFormat(
    name="ISO_YEAR_MONTH_DAY",
    regex_pattern=r"^\d{4}\-\d{2}\-\d{2}$",
    arrow_pattern="YYYY-MM-DD",
    timeframe="day",
)
ISO_YEAR_WEEK = IsoDateFormat(
    name="ISO_YEAR_WEEK",
    regex_pattern=r"^\d{4}\-{0,1}W\d{2}$",
    arrow_pattern="W",
    timeframe="week",
)


@dataclass
class SsbDateFormat(DateFormat):
    """A subclass of Dateformat with relevant patterns for SSB unique dates."""

    ssb_dates: dict

    def get_floor(self, period_string: str) -> date | None:
        """Convert SSB format to date-string and return first date.

        If not existing SSB format, return None

        >>> SSB_BIMESTER.get_floor("2003B8")
        None

        """
        try:
            year = period_string[:4]
            month = self.ssb_dates[period_string[-2:]]["start"]
            period = year + month
            return arrow.get(period, self.arrow_pattern).floor(self.timeframe).date()
        except KeyError:
            logger.exception("Error while converting to SSB date format")
            return None

    def get_ceil(self, period_string: str) -> date | None:
        """Convert SSB format to date-string and return last date.

        If not excisting SSB format, return None

        >>> SSB_TRIANNUAL.get_ceil("1999T11")
        None

        """
        try:
            year = period_string[:4]
            month = self.ssb_dates[period_string[-2:]]["end"]
            period = year + month
            return arrow.get(period, self.arrow_pattern).ceil(self.timeframe).date()
        except KeyError:
            return None


SSB_BIMESTER = SsbDateFormat(
    name="SSB_BIMESTER",
    regex_pattern=r"^\d{4}[B]\d{1}$",
    arrow_pattern="YYYYMM",
    timeframe="month",
    ssb_dates={
        "B1": {
            "start": "01",
            "end": "02",
        },
        "B2": {
            "start": "03",
            "end": "04",
        },
        "B3": {
            "start": "05",
            "end": "06",
        },
        "B4": {
            "start": "07",
            "end": "08",
        },
        "B5": {
            "start": "09",
            "end": "10",
        },
        "B6": {
            "start": "11",
            "end": "12",
        },
    },
)

SSB_QUARTERLY = SsbDateFormat(
    name="SSB_QUARTERLY",
    regex_pattern=r"^\d{4}[Q]\d{1}$",
    arrow_pattern="YYYYMM",
    timeframe="month",
    ssb_dates={
        "Q1": {
            "start": "01",
            "end": "03",
        },
        "Q2": {
            "start": "04",
            "end": "06",
        },
        "Q3": {
            "start": "07",
            "end": "09",
        },
        "Q4": {
            "start": "10",
            "end": "12",
        },
    },
)

SSB_TRIANNUAL = SsbDateFormat(
    name="SSB_TRIANNUAL",
    regex_pattern=r"^\d{4}[T]\d{1}$",
    arrow_pattern="YYYYMM",
    timeframe="month",
    ssb_dates={
        "T1": {
            "start": "01",
            "end": "04",
        },
        "T2": {
            "start": "05",
            "end": "08",
        },
        "T3": {
            "start": "09",
            "end": "12",
        },
    },
)
SSB_HALF_YEAR = SsbDateFormat(
    name="SSB_HALF_YEAR",
    regex_pattern=r"^\d{4}[H]\d{1}$",
    arrow_pattern="YYYYMM",
    timeframe="month",
    ssb_dates={
        "H1": {
            "start": "01",
            "end": "06",
        },
        "H2": {
            "start": "07",
            "end": "12",
        },
    },
)

SUPPORTED_DATE_FORMATS: list[IsoDateFormat | SsbDateFormat] = [
    ISO_YEAR,
    ISO_YEAR_MONTH,
    ISO_YEAR_MONTH_DAY,
    ISO_YEAR_WEEK,
    SSB_BIMESTER,
    SSB_QUARTERLY,
    SSB_TRIANNUAL,
    SSB_HALF_YEAR,
]


def categorize_period_string(period: str) -> IsoDateFormat | SsbDateFormat:
    """Categorize a period string into one of the supported date formats.

    If the period string is not recognized, a NotImplementedError is raised.

    Examples:
    >>> date_format = categorize_period_string('2022-W01')
    >>> date_format.name
    ISO_YEAR_WEEK

    >>> date_format = categorize_period_string('1954T2')
    >>> date_format.name
    SSB_TRIANNUAL

    >>> categorize_period_string('unknown format')
    Traceback (most recent call last):
     ...
    NotImplementedError: Period format unknown format is not supported
    """
    for date_format in SUPPORTED_DATE_FORMATS:
        if re.match(date_format.regex_pattern, period):
            return date_format

    msg = f"Period format {period} is not supported"
    raise NotImplementedError(
        msg,
    )


class DaplaDatasetPathInfo:
    """Extract info from a path following SSB's dataset naming convention."""

    def __init__(self, dataset_path: str | os.PathLike[str]) -> None:
        """Digest the path so that it's ready for further parsing."""
        self.dataset_path = pathlib.Path(dataset_path)
        self.dataset_name_sections = self.dataset_path.stem.split("_")
        self._period_strings = self._extract_period_strings(self.dataset_name_sections)

    @staticmethod
    def _extract_period_strings(dataset_name_sections: list[str]) -> list[str]:
        """Extract period strings from dataset name sections.

        Iterates over the dataset name sections and returns a list of strings
        that match the year regex, stripping the first character. This extracts
        the year periods from the dataset name.

        Examples:
        >>> DaplaDatasetPathInfo._extract_period_strings(['p2022', 'kommune', 'v1'])
        ['2022']

        >>> DaplaDatasetPathInfo._extract_period_strings(['p2022-01', 'p2023-06', 'kommune', 'v1'])
        ['2022-01', '2023-06']

        >>> DaplaDatasetPathInfo._extract_period_strings(['p1990Q1', 'kommune', 'v1'])
        ['1990Q1']

        >>> DaplaDatasetPathInfo._extract_period_strings(['varehandel','v1']) # No date will return empty string
        []

        """

        def insert_p(regex: str) -> str:
            r"""Insert a p as the second character.

            Examples:
            >>> insert_p(r"^\d{4}[H]\d{1}$")
            '^p\d{4}[H]\d{1}$'
            """
            return regex[:1] + "p" + regex[1:]

        return [
            x[1:]
            for x in dataset_name_sections
            if any(
                re.match(insert_p(date_format.regex_pattern), x)
                for date_format in SUPPORTED_DATE_FORMATS
            )
        ]

    def _extract_period_string_from_index(self, index: int) -> str | None:
        try:
            return self._period_strings[index]
        except IndexError:
            return None

    def _extract_norwegian_dataset_state_path_part(
        self,
        dataset_state: DataSetState,
    ) -> set:
        norwegian_dataset_state_path_part = dataset_state.get_value_for_language(
            SupportedLanguages.NORSK_BOKMÅL,
        ).lower()
        return {norwegian_dataset_state_path_part.replace(" ", x) for x in ["-", "_"]}

    @property
    def contains_data_from(self) -> datetime.date | None:
        """The earliest date from which data in the dataset is relevant for."""
        period_string = self._extract_period_string_from_index(0)
        if not period_string or (
            len(self._period_strings) > 1 and period_string > self._period_strings[1]
        ):
            return None
        date_format = categorize_period_string(period_string)
        return date_format.get_floor(period_string)

    @property
    def contains_data_until(self) -> datetime.date | None:
        """The latest date until which data in the dataset is relevant for."""
        first_period_string = self._extract_period_string_from_index(0)
        second_period_string = self._extract_period_string_from_index(1)
        period_string = second_period_string or first_period_string
        if not period_string or (
            second_period_string
            and first_period_string is not None
            and second_period_string < first_period_string
        ):
            return None
        date_format = categorize_period_string(period_string)
        return date_format.get_ceil(period_string)

    @property
    def dataset_state(
        self,
    ) -> DataSetState | None:
        """Extract the dataset state from the path.

        Examples:
        >>> DaplaDatasetPathInfo('klargjorte_data/person_data_v1.parquet').dataset_state
        <DataSetState.PROCESSED_DATA: 'PROCESSED_DATA'>
        >>> DaplaDatasetPathInfo('utdata/min_statistikk/person_data_v1.parquet').dataset_state
        <DataSetState.OUTPUT_DATA: 'OUTPUT_DATA'>
        >>> DaplaDatasetPathInfo('my_special_data/person_data_v1.parquet').dataset_state
        None
        """
        dataset_path_parts = set(self.dataset_path.parts)
        for s in DataSetState:
            # We assume that files are saved in the Norwegian language as specified by SSB.
            norwegian_dataset_state_path_part_variations = (
                self._extract_norwegian_dataset_state_path_part(s)
            )
            # Match on any of the variations anywhere in the path.
            if norwegian_dataset_state_path_part_variations.intersection(
                dataset_path_parts,
            ):
                return s
        return None

    @property
    def dataset_version(
        self,
    ) -> str | None:
        """Extract version information if exists in filename.

        Examples:
        >>> DaplaDatasetPathInfo('person_data_v1.parquet').dataset_version
        '1'
        >>> DaplaDatasetPathInfo('person_data_v20.parquet').dataset_version
        '20'
        >>> DaplaDatasetPathInfo('person_data.parquet').dataset_version
        None
        """
        minimum_elements_in_file_name: Final[int] = 2
        minimum_characters_in_version_string: Final[int] = 2
        if len(self.dataset_name_sections) >= minimum_elements_in_file_name:
            last_filename_element = str(self.dataset_name_sections[-1])
            if (
                len(last_filename_element) >= minimum_characters_in_version_string
                and last_filename_element[0:1] == "v"
                and last_filename_element[1:].isdigit()
            ):
                return last_filename_element[1:]
        return None

    @property
    def statistic_short_name(
        self,
    ) -> str | None:
        """Extract the statistical short name from the filepath right before the dataset_state, based on the Dapla filepath naming convention.

        Examples:
        >>> DaplaDatasetPathInfo('prosjekt/befolkning/klargjorte_data/person_data_v1.parquet').statistic_short_name
        befolkning
        >>> DaplaDatasetPathInfo('befolkning/inndata/person_data_v1.parquet').statistic_short_name
        befolkning
        >>> DaplaDatasetPathInfo('befolkning/person_data.parquet').statistic_short_name
        None
        """
        dataset_state = self.dataset_state
        if dataset_state is not None:
            dataset_state_names = self._extract_norwegian_dataset_state_path_part(
                dataset_state,
            )
            dataset_path_parts = list(self.dataset_path.parts)
            for i in dataset_state_names:
                if i in dataset_path_parts and dataset_path_parts.index(i) != 0:
                    return dataset_path_parts[dataset_path_parts.index(i) - 1]
        return None



"""Functionality common to displaying dataset and variables metadata."""

from __future__ import annotations

import logging
import re
import typing as t
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any

import dash_bootstrap_components as dbc
import ssb_dash_components as ssb
from dash import dcc

from datadoc import state
from datadoc.enums import SupportedLanguages
from datadoc.frontend.callbacks.utils import get_language_strings_enum

if TYPE_CHECKING:
    from collections.abc import Callable
    from enum import Enum

    from dash.development.base_component import Component
    from datadoc_model import model
    from datadoc_model.model import LanguageStringType
    from pydantic import BaseModel

    from datadoc.frontend.callbacks.utils import MetadataInputTypes

logger = logging.getLogger(__name__)

DATASET_METADATA_INPUT = "dataset-metadata-input"

VARIABLES_METADATA_INPUT = "variables-metadata-input"
VARIABLES_METADATA_DATE_INPUT = "variables-metadata-date-input"
DATASET_METADATA_DATE_INPUT = "dataset-metadata-date-input"


def get_enum_options_for_language(
    enum: Enum,
    language: SupportedLanguages,
) -> list[dict[str, str]]:
    """Generate the list of options based on the currently chosen language."""
    dropdown_options = [
        {
            "title": i.get_value_for_language(language),
            "id": i.name,
        }
        for i in get_language_strings_enum(enum)  # type: ignore [attr-defined]
    ]
    dropdown_options.insert(0, {"title": "", "id": ""})
    return dropdown_options


def empty_kwargs_factory() -> dict[str, t.Any]:
    """Initialize the field extra_kwargs.

    We aren't allowed to directly assign a mutable type like a dict to
    a dataclass field.
    """
    return {}


def get_standard_metadata(metadata: BaseModel, identifier: str) -> MetadataInputTypes:
    """Get a metadata value from the model."""
    return getattr(metadata, identifier)


def get_metadata_and_stringify(metadata: BaseModel, identifier: str) -> str | None:
    """Get a metadata value from the model and cast to string."""
    value = get_standard_metadata(metadata, identifier)
    if value is None:
        return None
    return str(value)


def get_date_metadata_and_stringify(metadata: BaseModel, identifier: str) -> str | None:
    """Get a metadata date value from the model.

    Handle converting datetime format to date format string.
    """
    value = get_standard_metadata(metadata, identifier)
    if value is None:
        return ""
    logger.info("Date registered: %s", value)
    date = str(value)
    # Pattern for datetime without T, with space - used for variables
    pattern = r"\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}\+\d{2}:\d{2}"
    if re.match(pattern, date):
        convert_date_to_iso = date.replace(" ", "T")
        logger.info("Date converted to iso format: %s", convert_date_to_iso)
        convert_date_format = convert_date_to_iso[:10]
        logger.info("Display date: %s", convert_date_format)
        return convert_date_format
    convert_value = date[:10]
    logger.info("Display date: %s", convert_value)
    return convert_value


def get_multi_language_metadata(metadata: BaseModel, identifier: str) -> str | None:
    """Get a metadata value supporting multiple languages from the model."""
    value: LanguageStringType | None = getattr(metadata, identifier)
    if value is None:
        return value
    return str(getattr(value, state.current_metadata_language))


def get_comma_separated_string(metadata: BaseModel, identifier: str) -> str:
    """Get a metadata value which is a list of strings from the model and convert it to a comma separated string."""
    value: list[str] = getattr(metadata, identifier)
    try:
        return ", ".join(value)
    except TypeError:
        logger.exception("Type error")
        return ""


@dataclass
class DisplayMetadata:
    """Controls for how a given metadata field should be displayed."""

    identifier: str
    display_name: str
    description: str
    obligatory: bool = False
    editable: bool = True
    url: bool = False
    multiple_language_support: bool = False


@dataclass
class MetadataInputField(DisplayMetadata):
    """Controls how a input field should be displayed."""

    type: str = "text"
    extra_kwargs: dict[str, Any] = field(default_factory=empty_kwargs_factory)
    value_getter: Callable[[BaseModel, str], Any] = get_metadata_and_stringify

    def render(
        self,
        component_id: dict,
        language: str,  # noqa: ARG002 Required by Dash
        metadata: BaseModel,
    ) -> ssb.Input:
        """Build component."""
        value = self.value_getter(metadata, self.identifier)
        return ssb.Input(
            label=self.display_name,
            id=component_id,
            debounce=True,
            type=self.type,
            disabled=not self.editable,
            value=value,
            className="input-component",
        )


@dataclass
class MetadataDropdownField(DisplayMetadata):
    """Control how a Dropdown should be displayed."""

    extra_kwargs: dict[str, Any] = field(default_factory=empty_kwargs_factory)
    value_getter: Callable[[BaseModel, str], Any] = get_metadata_and_stringify
    # fmt: off
    options_getter: Callable[[SupportedLanguages], list[dict[str, str]]] = lambda _: []  # noqa: E731
    # fmt: on

    def render(
        self,
        component_id: dict,
        language: str,
        dataset: BaseModel,
    ) -> ssb.Dropdown:
        """Build Dropdown component."""
        value = self.value_getter(dataset, self.identifier)
        return ssb.Dropdown(
            header=self.display_name,
            id=component_id,
            items=self.options_getter(SupportedLanguages(language)),
            value=value,
            className="dropdown-component",
        )


@dataclass
class MetadataPeriodField(DisplayMetadata):
    """Control how fields which define a time period are displayed for Dataset.

    These are a special case since two fields have a relationship to one another.>
    """

    id_type: str = ""
    extra_kwargs: dict[str, Any] = field(default_factory=empty_kwargs_factory)
    value_getter: Callable[[BaseModel, str], Any] = get_date_metadata_and_stringify
    type: str = "date"

    def render(
        self,
        component_id: dict,
        language: str,  # noqa: ARG002 Required by Dash
        dataset: BaseModel,
    ) -> ssb.Input:
        """Build Input date component."""
        value = self.value_getter(dataset, self.identifier)
        component_id["type"] = self.id_type
        return ssb.Input(
            label=self.display_name,
            id=component_id,
            debounce=False,
            type=self.type,
            disabled=not self.editable,
            value=value,
            className="input-component",
        )


@dataclass
class MetadataCheckboxField(DisplayMetadata):
    """Controls for how a checkbox metadata field should be displayed."""

    extra_kwargs: dict[str, Any] = field(default_factory=empty_kwargs_factory)
    value_getter: Callable[[BaseModel, str], Any] = get_standard_metadata

    def render(
        self,
        variable_id: dict,
        language: str,  # noqa: ARG002 Required by Dash
        variable: model.Variable,
    ) -> dbc.Checkbox:
        """Build Checkbox component."""
        value = self.value_getter(variable, self.identifier)
        return dbc.Checkbox(
            label=self.display_name,
            id=variable_id,
            disabled=not self.editable,
            label_class_name="ssb-checkbox checkbox-label",
            class_name="ssb-checkbox",
            value=value,
        )


VariablesFieldTypes = (
    MetadataInputField
    | MetadataDropdownField
    | MetadataCheckboxField
    | MetadataPeriodField
)

DatasetFieldTypes = MetadataInputField | MetadataDropdownField | MetadataPeriodField


@dataclass
class DisplayDatasetMetadata(DisplayMetadata):
    """Controls for how a given metadata field should be displayed.

    Specific to dataset fields.
    """

    extra_kwargs: dict[str, Any] = field(default_factory=empty_kwargs_factory)
    component: type[Component] = dcc.Input
    value_getter: Callable[[BaseModel, str], Any] = get_metadata_and_stringify


@dataclass
class DisplayDatasetMetadataDropdown(DisplayDatasetMetadata):
    """Include the possible options which a user may choose from."""

    # fmt: off
    options_getter: Callable[[SupportedLanguages], list[dict[str, str]]] = lambda _: []  # noqa: E731
    # fmt: on
    extra_kwargs: dict[str, Any] = field(default_factory=empty_kwargs_factory)
    component: type[Component] = dcc.Dropdown



"""General utilities."""

import datetime
import importlib
from typing import Any

from datadoc_model import model
from pydantic import AnyUrl

from datadoc.enums import SupportedLanguages

METADATA_DOCUMENT_FILE_SUFFIX = "__DOC.json"


def running_in_notebook() -> bool:
    """Return True if running in Jupyter Notebook."""
    try:
        return bool(get_ipython().__class__.__name__ == "ZMQInteractiveShell")  # type: ignore [name-defined]
    except NameError:
        # The get_ipython method is globally available in ipython interpreters
        # as used in Jupyter. However it is not available in other python
        # interpreters and will throw a NameError. Therefore we're not running
        # in Jupyter.
        return False


def calculate_percentage(completed: int, total: int) -> int:
    """Calculate percentage as a rounded integer."""
    return round((completed / total) * 100)


def get_display_values(
    variable: model.Variable,
    current_language: SupportedLanguages,
) -> dict[str, Any]:
    """Return a dictionary representation of Model.DataDocVariable with strings in the currently selected language."""
    return_dict = {}
    for field_name, value in variable:
        if isinstance(value, model.LanguageStringType):
            return_dict[field_name] = value.model_dump()[current_language.value]
        elif isinstance(value, AnyUrl):
            return_dict[field_name] = str(value)
        else:
            return_dict[field_name] = value
    return return_dict


def pick_random_port() -> int:
    """Pick a random free port number.

    The function will bind a socket to port 0, and a random free port from
    1024 to 65535 will be selected by the operating system.
    """
    import socket

    with socket.socket() as sock:
        sock.bind(("", 0))
        return int(sock.getsockname()[1])


def get_timestamp_now() -> datetime.datetime:
    """Return a timestamp for the current moment."""
    return datetime.datetime.now(tz=datetime.timezone.utc)


def get_app_version() -> str:
    """Get the version of the Datadoc package."""
    return importlib.metadata.distribution("ssb-datadoc").version



"""Components and layout which are not inside a tab."""

from __future__ import annotations

import dash_bootstrap_components as dbc
import ssb_dash_components as ssb
from dash import dcc
from dash import html

from datadoc import state
from datadoc.enums import SupportedLanguages
from datadoc.frontend.callbacks.utils import get_dataset_path
from datadoc.utils import get_app_version

COLORS = {"dark_1": "#F0F8F9", "green_1": "#ECFEED", "green_4": "#00824D"}

header = ssb.Header(
    [ssb.Title("DataDoc", size=1, id="main-title", className="main-title")],
    className="datadoc-header",
)

progress_bar = dbc.CardBody(
    children=[dbc.Progress(id="progress-bar", color=COLORS["green_4"], value=40)],
    className="progress-bar-wrapper",
)


def build_language_dropdown() -> dbc.Row:
    """Build the language dropdown."""
    return dbc.CardBody(
        dbc.Row(
            [
                dbc.Col(
                    html.P(f"v{get_app_version()}", className="small"),
                    align="end",
                ),
                dbc.Col(
                    dcc.Dropdown(
                        id="language-dropdown",
                        searchable=False,
                        value=state.current_metadata_language.value,
                        className="ssb-dropdown",
                        options=[
                            {"label": i.name, "value": i.value}
                            for i in SupportedLanguages
                        ],
                    ),
                    align="center",
                    width="auto",
                ),
            ],
            justify="between",
        ),
    )


def build_controls_bar() -> html.Section:
    """Build the Controls Bar.

    This contains:
    - A text input to specify the path to a dataset
    - A button to open a dataset
    - A button to save metadata to disk
    """
    return html.Section(
        [
            ssb.Input(
                label="Filsti",
                value=get_dataset_path(),
                className="file-path-input",
                id="dataset-path-input",
            ),
            ssb.Button(
                children=["Åpne fil"],
                id="open-button",
                className="file-open-button",
            ),
            ssb.Button(
                children=["Lagre metadata"],
                id="save-button",
                className="file-save-button",
            ),
        ],
        className="control-bar-section",
    )



"""All decorated callback functions should be defined here.

Implementations of the callback functionality should be in other functions (in other files), to enable unit testing.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from dash import ALL
from dash import MATCH
from dash import Dash
from dash import Input
from dash import Output
from dash import State
from dash import ctx
from dash import no_update

from datadoc import state
from datadoc.enums import SupportedLanguages
from datadoc.frontend.callbacks.dataset import accept_dataset_metadata_date_input
from datadoc.frontend.callbacks.dataset import accept_dataset_metadata_input
from datadoc.frontend.callbacks.dataset import open_dataset_handling
from datadoc.frontend.callbacks.utils import update_global_language_state
from datadoc.frontend.callbacks.variables import accept_variable_metadata_date_input
from datadoc.frontend.callbacks.variables import accept_variable_metadata_input
from datadoc.frontend.components.builders import build_dataset_edit_section
from datadoc.frontend.components.builders import build_edit_section
from datadoc.frontend.components.builders import build_ssb_accordion
from datadoc.frontend.components.dataset_tab import SECTION_WRAPPER_ID
from datadoc.frontend.components.variables_tab import ACCORDION_WRAPPER_ID
from datadoc.frontend.components.variables_tab import VARIABLES_INFORMATION_ID
from datadoc.frontend.fields.display_base import DATASET_METADATA_DATE_INPUT
from datadoc.frontend.fields.display_base import DATASET_METADATA_INPUT
from datadoc.frontend.fields.display_base import VARIABLES_METADATA_DATE_INPUT
from datadoc.frontend.fields.display_base import VARIABLES_METADATA_INPUT
from datadoc.frontend.fields.display_dataset import NON_EDITABLE_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import OBLIGATORY_EDITABLE_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import OPTIONAL_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import DatasetIdentifiers
from datadoc.frontend.fields.display_variables import OBLIGATORY_VARIABLES_METADATA
from datadoc.frontend.fields.display_variables import OPTIONAL_VARIABLES_METADATA
from datadoc.frontend.fields.display_variables import VariableIdentifiers

if TYPE_CHECKING:
    import dash_bootstrap_components as dbc

    from datadoc.frontend.callbacks.utils import MetadataInputTypes

logger = logging.getLogger(__name__)


def register_callbacks(app: Dash) -> None:
    """Define and register callbacks."""

    @app.callback(
        Output("progress-bar", "value"),
        Output("progress-bar", "label"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        Input(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": ALL,
                "id": ALL,
            },
            "value",
        ),
    )
    def callback_update_progress(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
        data: list[dict],  # noqa: ARG001 argument required by Dash
    ) -> tuple[int, str]:
        """Update the progress bar when new data is entered."""
        completion = state.metadata.percent_complete
        return completion, f"{completion}%"

    @app.callback(
        Output("saved-metadata-success", "is_open"),
        Input("save-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def callback_save_metadata_file(n_clicks: int) -> bool:
        """Save the metadata document to disk."""
        if n_clicks and n_clicks > 0:
            state.metadata.write_metadata_document()
            return True

        return False

    @app.callback(
        Output("dataset-validation-error", "is_open"),
        Output("dataset-validation-explanation", "children"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> tuple[bool, str]:
        """Save updated dataset metadata values.

        Will display an alert if validation fails.
        """
        # Get the ID of the input that changed. This MUST match the attribute name defined in DataDocDataSet
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["id"],
        )

    @app.callback(
        Output("opened-dataset-success", "is_open"),
        Output("opened-dataset-error", "is_open"),
        Output("opened-dataset-error-explanation", "children"),
        Output("language-dropdown", "value"),  # Used to force reload of metadata
        Input("open-button", "n_clicks"),
        State("dataset-path-input", "value"),
    )
    def callback_open_dataset(
        n_clicks: int,
        dataset_path: str,
    ) -> tuple[bool, bool, str, str]:
        """Open a dataset.

        Shows an alert on success or failure.

        To trigger reload of data in the UI, we update the
        language dropdown. This is a hack and could be replaced
        by a more formal mechanism.
        """
        return open_dataset_handling(n_clicks, dataset_path)

    @app.callback(
        Output(VARIABLES_INFORMATION_ID, "children"),
        Input("language-dropdown", "value"),
        prevent_initial_call=True,
    )
    def callback_populate_variables_info_section(
        language: str,  # noqa: ARG001 Dash requires arguments for all Inputs
    ) -> str:
        return f"Datasettet inneholder {len(state.metadata.variables)} variabler."

    @app.callback(
        Output(ACCORDION_WRAPPER_ID, "children"),
        Input("language-dropdown", "value"),
        prevent_initial_call=True,
    )
    def callback_populate_variables_workspace(
        language: str,
    ) -> list:
        """Create variable workspace with accordions for variables."""
        update_global_language_state(SupportedLanguages(language))
        logger.info("Populating new variables workspace")
        return [
            build_ssb_accordion(
                variable.short_name,
                {
                    "type": "variables-accordion",
                    "id": f"{variable.short_name}-{language}",  # Insert language into the ID to invalidate browser caches
                },
                variable.short_name,
                children=[
                    build_edit_section(
                        OBLIGATORY_VARIABLES_METADATA,
                        "Obligatorisk",
                        variable,
                        state.current_metadata_language.value,
                    ),
                    build_edit_section(
                        OPTIONAL_VARIABLES_METADATA,
                        "Anbefalt",
                        variable,
                        state.current_metadata_language.value,
                    ),
                ],
            )
            for variable in list(state.metadata.variables)
        ]

    @app.callback(
        Output(SECTION_WRAPPER_ID, "children"),
        Input("language-dropdown", "value"),
        Input("open-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def callback_populate_dataset_workspace(
        language: str,
        n_clicks: int,
    ) -> list:
        """Create dataset workspace with sections."""
        update_global_language_state(SupportedLanguages(language))
        logger.info("Populating new dataset workspace")
        if n_clicks:
            return [
                build_dataset_edit_section(
                    "Obligatorisk",
                    OBLIGATORY_EDITABLE_DATASET_METADATA,
                    state.current_metadata_language,
                    state.metadata.dataset,
                    {"type": "dataset-edit-section", "id": f"obligatory-{language}"},
                ),
                build_dataset_edit_section(
                    "Anbefalt",
                    OPTIONAL_DATASET_METADATA,
                    state.current_metadata_language,
                    state.metadata.dataset,
                    {"type": "dataset-edit-section", "id": f"recommended-{language}"},
                ),
                build_dataset_edit_section(
                    "Maskingenerert",
                    NON_EDITABLE_DATASET_METADATA,
                    state.current_metadata_language,
                    state.metadata.dataset,
                    {"type": "dataset-edit-section", "id": f"machine-{language}"},
                ),
            ]
        return no_update

    @app.callback(
        Output(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_variable_metadata_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> dbc.Alert:
        """Save updated variable metadata values."""
        message = accept_variable_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["variable_short_name"],
            ctx.triggered_id["id"],
        )
        if not message:
            # No error to display.
            return False, ""

        return True, message

    @app.callback(
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "errorMessage",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "value",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_variable_metadata_date_input(
        contains_data_from: str,
        contains_data_until: str,
    ) -> dbc.Alert:
        """Special case handling for date fields which have a relationship to one another."""
        return accept_variable_metadata_date_input(
            VariableIdentifiers(ctx.triggered_id["id"]),
            ctx.triggered_id["variable_short_name"],
            contains_data_from,
            contains_data_until,
        )

    @app.callback(
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "errorMessage",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "value",
        ),
        Input(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_date_input(
        contains_data_from: str,
        contains_data_until: str,
    ) -> dbc.Alert:
        """Special case handling for date fields which have a relationship to one another."""
        return accept_dataset_metadata_date_input(
            DatasetIdentifiers(ctx.triggered_id["id"]),
            contains_data_from,
            contains_data_until,
        )



"""Functionality for displaying dataset metadata."""

from __future__ import annotations

import functools
import logging
from enum import Enum
from typing import TYPE_CHECKING

from datadoc import enums
from datadoc import state
from datadoc.frontend.fields.display_base import DATASET_METADATA_DATE_INPUT
from datadoc.frontend.fields.display_base import DatasetFieldTypes
from datadoc.frontend.fields.display_base import DisplayDatasetMetadataDropdown
from datadoc.frontend.fields.display_base import MetadataDropdownField
from datadoc.frontend.fields.display_base import MetadataInputField
from datadoc.frontend.fields.display_base import MetadataPeriodField
from datadoc.frontend.fields.display_base import get_comma_separated_string
from datadoc.frontend.fields.display_base import get_enum_options_for_language
from datadoc.frontend.fields.display_base import get_metadata_and_stringify
from datadoc.frontend.fields.display_base import get_multi_language_metadata

if TYPE_CHECKING:
    from datadoc.enums import SupportedLanguages

logger = logging.getLogger(__name__)


def get_statistical_subject_options(
    language: SupportedLanguages,
) -> list[dict[str, str]]:
    """Generate the list of options for statistical subject."""
    dropdown_options = [
        {
            "title": f"{primary.get_title(language)} - {secondary.get_title(language)}",
            "id": secondary.subject_code,
        }
        for primary in state.statistic_subject_mapping.primary_subjects
        for secondary in primary.secondary_subjects
    ]
    dropdown_options.insert(0, {"title": "", "id": ""})
    return dropdown_options


def get_unit_type_options(
    language: SupportedLanguages,
) -> list[dict[str, str]]:
    """Collect the unit type options for the given language."""
    dropdown_options = [
        {
            "title": unit_type.get_title(language),
            "id": unit_type.code,
        }
        for unit_type in state.unit_types.classifications
    ]
    dropdown_options.insert(0, {"title": "", "id": ""})
    return dropdown_options


def get_owner_options(
    language: SupportedLanguages,
) -> list[dict[str, str]]:
    """Collect the owner options for the given language."""
    dropdown_options = [
        {
            "title": f"{option.code} - {option.get_title(language)}",
            "id": option.code,
        }
        for option in state.organisational_units.classifications
    ]
    dropdown_options.insert(0, {"title": "", "id": ""})
    return dropdown_options


class DatasetIdentifiers(str, Enum):
    """As defined here: https://statistics-norway.atlassian.net/l/c/aoSfEWJU."""

    SHORT_NAME = "short_name"
    ASSESSMENT = "assessment"
    DATASET_STATUS = "dataset_status"
    DATASET_STATE = "dataset_state"
    NAME = "name"
    DESCRIPTION = "description"
    DATA_SOURCE = "data_source"
    REGISTER_URI = "register_uri"
    POPULATION_DESCRIPTION = "population_description"
    VERSION = "version"
    VERSION_DESCRIPTION = "version_description"
    UNIT_TYPE = "unit_type"
    TEMPORALITY_TYPE = "temporality_type"
    SUBJECT_FIELD = "subject_field"
    KEYWORD = "keyword"
    SPATIAL_COVERAGE_DESCRIPTION = "spatial_coverage_description"
    ID = "id"
    OWNER = "owner"
    FILE_PATH = "file_path"
    METADATA_CREATED_DATE = "metadata_created_date"
    METADATA_CREATED_BY = "metadata_created_by"
    METADATA_LAST_UPDATED_DATE = "metadata_last_updated_date"
    METADATA_LAST_UPDATED_BY = "metadata_last_updated_by"
    CONTAINS_DATA_FROM = "contains_data_from"
    CONTAINS_DATA_UNTIL = "contains_data_until"


DISPLAY_DATASET: dict[
    DatasetIdentifiers,
    DatasetFieldTypes,
] = {
    DatasetIdentifiers.SHORT_NAME: MetadataInputField(
        identifier=DatasetIdentifiers.SHORT_NAME.value,
        display_name="Kortnavn",
        description="Navn på (fysisk) datafil, datatabell eller datasett",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.ASSESSMENT: MetadataDropdownField(
        identifier=DatasetIdentifiers.ASSESSMENT.value,
        display_name="Verdivurdering",
        description="Verdivurdering (sensitivitetsklassifisering) for datasettet.",
        options_getter=functools.partial(
            get_enum_options_for_language,
            enums.Assessment,
        ),
    ),
    DatasetIdentifiers.DATASET_STATUS: MetadataDropdownField(
        identifier=DatasetIdentifiers.DATASET_STATUS.value,
        display_name="Status",
        description="Livssyklus for datasettet",
        options_getter=functools.partial(
            get_enum_options_for_language,
            enums.DataSetStatus,
        ),
        obligatory=True,
    ),
    DatasetIdentifiers.DATASET_STATE: MetadataDropdownField(
        identifier=DatasetIdentifiers.DATASET_STATE.value,
        display_name="Datatilstand",
        description="Datatilstand. Se Intern dokument 2021- 17  Datatilstander i SSB",
        obligatory=True,
        options_getter=functools.partial(
            get_enum_options_for_language,
            enums.DataSetState,
        ),
    ),
    DatasetIdentifiers.NAME: MetadataInputField(
        identifier=DatasetIdentifiers.NAME.value,
        display_name="Navn",
        description="Datasettnavn",
        obligatory=True,
        multiple_language_support=True,
        value_getter=get_metadata_and_stringify,
    ),
    DatasetIdentifiers.DATA_SOURCE: MetadataInputField(
        identifier=DatasetIdentifiers.DATA_SOURCE.value,
        display_name="Datakilde",
        description="Datakilde. Settes enten for datasettet eller variabelforekomst.",
        obligatory=True,
        multiple_language_support=True,
    ),
    DatasetIdentifiers.REGISTER_URI: MetadataInputField(
        identifier=DatasetIdentifiers.REGISTER_URI.value,
        display_name="Register URI",
        description="Lenke (URI) til register i registeroversikt (oversikt over alle registre meldt Datatilsynet (oppdatering foretas av sikkerhetsrådgiver))",
        multiple_language_support=True,
        url=True,
        type="url",
    ),
    DatasetIdentifiers.POPULATION_DESCRIPTION: MetadataInputField(
        identifier=DatasetIdentifiers.POPULATION_DESCRIPTION.value,
        display_name="Populasjon",
        description="Populasjonen datasettet dekker. Populasjonsbeskrivelsen inkluderer enhetstype, geografisk dekningsområde og tidsperiode.",
        obligatory=True,
        multiple_language_support=True,
    ),
    DatasetIdentifiers.VERSION: MetadataInputField(
        identifier=DatasetIdentifiers.VERSION.value,
        display_name="Versjon",
        description="Versjon",
        extra_kwargs={"type": "number", "min": 1},
        obligatory=True,
    ),
    DatasetIdentifiers.VERSION_DESCRIPTION: MetadataInputField(
        identifier=DatasetIdentifiers.VERSION_DESCRIPTION.value,
        display_name="Versjonsbeskrivelse",
        description="Årsak/grunnlag for denne versjonen av datasettet i form av beskrivende tekst.",
        multiple_language_support=True,
        obligatory=True,
    ),
    DatasetIdentifiers.UNIT_TYPE: MetadataDropdownField(
        identifier=DatasetIdentifiers.UNIT_TYPE.value,
        display_name="Enhetstype",
        description="Primær enhetstype for datafil, datatabell eller datasett. Se  Vi jobber med en avklaring av behov for flere enhetstyper her.",
        multiple_language_support=False,
        options_getter=get_unit_type_options,
        obligatory=True,
    ),
    DatasetIdentifiers.TEMPORALITY_TYPE: MetadataDropdownField(
        identifier=DatasetIdentifiers.TEMPORALITY_TYPE.value,
        display_name="Temporalitetstype",
        description="Temporalitetstype. Settes enten for variabelforekomst eller datasett. Se Temporalitet, hendelser og forløp.",
        options_getter=functools.partial(
            get_enum_options_for_language,
            enums.TemporalityTypeType,
        ),
        obligatory=True,
    ),
    DatasetIdentifiers.DESCRIPTION: MetadataInputField(
        identifier=DatasetIdentifiers.DESCRIPTION.value,
        display_name="Beskrivelse",
        description="Beskrivelse av datasettet",
        multiple_language_support=True,
        obligatory=True,
    ),
    DatasetIdentifiers.SUBJECT_FIELD: MetadataDropdownField(
        identifier=DatasetIdentifiers.SUBJECT_FIELD.value,
        display_name="Statistikkområde",
        description="Primær statistikkområdet som datasettet inngår i",
        obligatory=True,
        # TODO @mmwinther: Remove multiple_language_support once the model is updated.
        # https://github.com/statisticsnorway/ssb-datadoc-model/issues/41
        multiple_language_support=True,
        options_getter=get_statistical_subject_options,
    ),
    DatasetIdentifiers.KEYWORD: MetadataInputField(
        identifier=DatasetIdentifiers.KEYWORD.value,
        display_name="Nøkkelord",
        description="En kommaseparert liste med søkbare nøkkelord som kan bidra til utvikling av effektive filtrerings- og søketjeneste.",
        value_getter=get_comma_separated_string,
    ),
    DatasetIdentifiers.SPATIAL_COVERAGE_DESCRIPTION: MetadataInputField(
        identifier=DatasetIdentifiers.SPATIAL_COVERAGE_DESCRIPTION.value,
        display_name="Geografisk dekningsområde",
        description="Beskrivelse av datasettets geografiske dekningsområde. Målet er på sikt at dette skal hentes fra Klass, men fritekst vil også kunne brukes.",
        multiple_language_support=True,
    ),
    DatasetIdentifiers.ID: MetadataInputField(
        identifier=DatasetIdentifiers.ID.value,
        display_name="ID",
        description="Unik SSB-identifikator for datasettet (løpenummer)",
        obligatory=True,
        editable=False,
        value_getter=get_metadata_and_stringify,
    ),
    DatasetIdentifiers.OWNER: MetadataDropdownField(
        identifier=DatasetIdentifiers.OWNER.value,
        display_name="Eier",
        description="Maskingenerert seksjonstilhørighet til den som oppretter metadata om datasettet, men kan korrigeres manuelt",
        obligatory=True,
        editable=True,
        multiple_language_support=False,
        options_getter=get_owner_options,
    ),
    DatasetIdentifiers.FILE_PATH: MetadataInputField(
        identifier=DatasetIdentifiers.FILE_PATH.value,
        display_name="Filsti",
        description="Filstien inneholder datasettets navn og stien til hvor det er lagret.",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.METADATA_CREATED_DATE: MetadataInputField(
        identifier=DatasetIdentifiers.METADATA_CREATED_DATE.value,
        display_name="Dato opprettet",
        description="Opprettet dato for metadata om datasettet",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.METADATA_CREATED_BY: MetadataInputField(
        identifier=DatasetIdentifiers.METADATA_CREATED_BY.value,
        display_name="Opprettet av",
        description="Opprettet av person. Kun til bruk i SSB.",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.METADATA_LAST_UPDATED_DATE: MetadataInputField(
        identifier=DatasetIdentifiers.METADATA_LAST_UPDATED_DATE.value,
        display_name="Dato oppdatert",
        description="Sist oppdatert dato for metadata om datasettet",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.METADATA_LAST_UPDATED_BY: MetadataInputField(
        identifier=DatasetIdentifiers.METADATA_LAST_UPDATED_BY.value,
        display_name="Oppdatert av",
        description="Siste endring utført av person. Kun til bruk i SSB.",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.CONTAINS_DATA_FROM: MetadataPeriodField(
        identifier=DatasetIdentifiers.CONTAINS_DATA_FROM.value,
        display_name="Inneholder data f.o.m.",
        description="ÅÅÅÅ-MM-DD",
        obligatory=True,
        editable=True,
        id_type=DATASET_METADATA_DATE_INPUT,
    ),
    DatasetIdentifiers.CONTAINS_DATA_UNTIL: MetadataPeriodField(
        identifier=DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
        display_name="Inneholder data t.o.m.",
        description="ÅÅÅÅ-MM-DD",
        obligatory=True,
        editable=True,
        id_type=DATASET_METADATA_DATE_INPUT,
    ),
}

for v in DISPLAY_DATASET.values():
    if v.multiple_language_support:
        v.value_getter = get_multi_language_metadata

MULTIPLE_LANGUAGE_DATASET_METADATA = [
    m.identifier for m in DISPLAY_DATASET.values() if m.multiple_language_support
]

OBLIGATORY_EDITABLE_DATASET_METADATA = [
    m for m in DISPLAY_DATASET.values() if m.obligatory and m.editable
]

OPTIONAL_DATASET_METADATA = [
    m for m in DISPLAY_DATASET.values() if not m.obligatory and m.editable
]

NON_EDITABLE_DATASET_METADATA = [m for m in DISPLAY_DATASET.values() if not m.editable]


# The order of this list MUST match the order of display components, as defined in DatasetTab.py
DISPLAYED_DATASET_METADATA: list[DatasetFieldTypes] = (
    OBLIGATORY_EDITABLE_DATASET_METADATA
    + OPTIONAL_DATASET_METADATA
    + NON_EDITABLE_DATASET_METADATA
)

DISPLAYED_DROPDOWN_DATASET_METADATA: list[DisplayDatasetMetadataDropdown] = [
    m
    for m in DISPLAYED_DATASET_METADATA
    if isinstance(m, DisplayDatasetMetadataDropdown)
]

OBLIGATORY_DATASET_METADATA_IDENTIFIERS: list[str] = [
    m.identifier for m in DISPLAY_DATASET.values() if m.obligatory and m.editable
]



"""Upgrade old metadata files to be compatible with new versions of the datadoc model.

An important principle of Datadoc is that we ALWAYS guarantee backwards compatibility of existing metadata documents.
This means that we guarantee that a user will never lose data, even if their document is decades old.

For each document version we release with breaking changes, we implement a handler and register the version by defining a
BackwardsCompatibleVersion instance. These documents will then be upgraded when they're opened in Datadoc.

A test must also be implemented for each new version.
"""

from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass
from datetime import datetime
from datetime import timezone
from typing import TYPE_CHECKING
from typing import Any

from datadoc_model.model import LanguageStringType

if TYPE_CHECKING:
    from collections.abc import Callable

VERSION_FIELD_NAME = "document_version"


class UnknownModelVersionError(Exception):
    """Throw this error if we haven't seen the version before."""

    def __init__(
        self,
        supplied_version: str,
        *args: tuple[Any, ...],
    ) -> None:
        """Initialize class."""
        super().__init__(args)
        self.supplied_version = supplied_version

    def __str__(self) -> str:
        """Return string representation."""
        return f"Document Version ({self.supplied_version}) of discovered file is not supported"


SUPPORTED_VERSIONS: OrderedDict[str, BackwardsCompatibleVersion] = OrderedDict()


@dataclass()
class BackwardsCompatibleVersion:
    """A version which we support with backwards compatibility."""

    version: str
    handler: Callable[[dict[str, Any]], dict[str, Any]]

    def __post_init__(self) -> None:
        """Register this version in the supported versions map."""
        SUPPORTED_VERSIONS[self.version] = self


def handle_current_version(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Nothing to do here."""
    return supplied_metadata


def handle_version_2_1_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for v2.1.0.

    Datatype changed from LanguageStringType to str for owner
    """
    data = supplied_metadata["dataset"]["owner"]
    supplied_metadata["dataset"]["owner"] = str(data["nb"] or data["nn"] or data["en"])
    supplied_metadata["document_version"] = "2.2.0"
    return supplied_metadata


def handle_version_1_0_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for v1.0.0."""
    datetime_fields = [("metadata_created_date"), ("metadata_last_updated_date")]
    for field in datetime_fields:
        if supplied_metadata["dataset"][field]:
            supplied_metadata["dataset"][field] = datetime.isoformat(
                datetime.fromisoformat(supplied_metadata["dataset"][field]).astimezone(
                    tz=timezone.utc,
                ),
                timespec="seconds",
            )
    if isinstance(supplied_metadata["dataset"]["data_source"], str):
        supplied_metadata["dataset"]["data_source"] = LanguageStringType(
            en=supplied_metadata["dataset"]["data_source"],
        )
    supplied_metadata["document_version"] = "2.1.0"
    return supplied_metadata


def handle_version_0_1_1(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for v0.1.1.

    PR ref: https://github.com/statisticsnorway/ssb-datadoc-model/pull/4.
    """
    key_renaming = [
        ("metadata_created_date", "created_date"),
        ("metadata_created_by", "created_by"),
        ("metadata_last_updated_date", "last_updated_date"),
        ("metadata_last_updated_by", "last_updated_by"),
    ]
    for new_key, old_key in key_renaming:
        supplied_metadata["dataset"][new_key] = supplied_metadata["dataset"].pop(
            old_key,
        )
    # Replace empty strings with None, empty strings are not valid for LanguageStrings values
    supplied_metadata["dataset"] = {
        k: None if v == "" else v for k, v in supplied_metadata["dataset"].items()
    }
    return supplied_metadata


# Register all the supported versions and their handlers.
# MUST be ordered from oldest to newest.
BackwardsCompatibleVersion(version="0.1.1", handler=handle_version_0_1_1)
BackwardsCompatibleVersion(version="1.0.0", handler=handle_version_1_0_0)
BackwardsCompatibleVersion(version="2.1.0", handler=handle_version_2_1_0)
BackwardsCompatibleVersion(version="2.2.0", handler=handle_current_version)


def upgrade_metadata(fresh_metadata: dict[str, Any]) -> dict[str, Any]:
    """Run the handler for this version to upgrade the document to the latest version."""
    # Special case for current version, we expose the current_model_version parameter for test purposes
    supplied_version = fresh_metadata[VERSION_FIELD_NAME]
    start_running_handlers = False
    # Run all the handlers in order from the supplied version onwards
    for k, v in SUPPORTED_VERSIONS.items():
        if k == supplied_version:
            start_running_handlers = True
        if start_running_handlers:
            fresh_metadata = v.handler(fresh_metadata)
    if not start_running_handlers:
        raise UnknownModelVersionError(supplied_version)
    return fresh_metadata



"""Callbacks relating to datasets."""

from __future__ import annotations

import logging
import traceback
from typing import TYPE_CHECKING

from pydantic import ValidationError

from datadoc import state
from datadoc.backend.datadoc_metadata import DataDocMetadata
from datadoc.enums import (
    SupportedLanguages,  # noqa: TCH001 import is needed for docs build
)
from datadoc.frontend.callbacks.utils import MetadataInputTypes
from datadoc.frontend.callbacks.utils import find_existing_language_string
from datadoc.frontend.callbacks.utils import get_dataset_path
from datadoc.frontend.callbacks.utils import parse_and_validate_dates
from datadoc.frontend.callbacks.utils import update_global_language_state
from datadoc.frontend.fields.display_dataset import DISPLAYED_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import DISPLAYED_DROPDOWN_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import MULTIPLE_LANGUAGE_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import DatasetIdentifiers
from datadoc.utils import METADATA_DOCUMENT_FILE_SUFFIX

if TYPE_CHECKING:
    from datadoc_model.model import LanguageStringType

logger = logging.getLogger(__name__)

VALIDATION_ERROR = "Validation error: "
DATE_VALIDATION_MESSAGE = f"{VALIDATION_ERROR}{DatasetIdentifiers.CONTAINS_DATA_FROM.value} must be the same or earlier date than {DatasetIdentifiers.CONTAINS_DATA_UNTIL.value}"


def open_file(file_path: str | None = None) -> DataDocMetadata:
    """Load the given dataset into a DataDocMetadata instance."""
    if file_path and file_path.endswith(METADATA_DOCUMENT_FILE_SUFFIX):
        logger.info("Opening existing metadata document %s", file_path)
        return DataDocMetadata(
            state.statistic_subject_mapping,
            metadata_document_path=file_path,
        )

    dataset = file_path or get_dataset_path()
    logger.info("Opening dataset %s", dataset)
    return DataDocMetadata(
        state.statistic_subject_mapping,
        dataset_path=str(dataset) if dataset else None,
    )


def open_dataset_handling(
    n_clicks: int,
    file_path: str,
) -> tuple[bool, bool, str, str]:
    """Handle errors and other logic around opening a dataset file."""
    if file_path:
        file_path = file_path.strip()

    try:
        state.metadata = open_file(file_path)
    except FileNotFoundError:
        logger.exception("File %s not found", str(file_path))
        return (
            False,
            True,
            f"Filen '{file_path}' finnes ikke.",
            state.current_metadata_language.value,
        )
    except Exception as e:  # noqa: BLE001
        return (
            False,
            True,
            "\n".join(traceback.format_exception_only(type(e), e)),
            state.current_metadata_language.value,
        )
    if n_clicks and n_clicks > 0:
        return True, False, "", state.current_metadata_language.value

    return False, False, "", state.current_metadata_language.value


def process_keyword(value: str) -> list[str]:
    """Convert a comma separated string to a list of strings.

    e.g. 'a,b ,c' -> ['a', 'b', 'c']
    """
    return [item.strip() for item in value.split(",")]


def process_special_cases(
    value: MetadataInputTypes | LanguageStringType,
    metadata_identifier: str,
) -> MetadataInputTypes | LanguageStringType:
    """Pre-process metadata where needed.

    Some types of metadata need processing before being saved
    to the model. Handle these cases here, other values are
    returned unchanged.
    """
    updated_value: MetadataInputTypes | LanguageStringType
    if metadata_identifier == DatasetIdentifiers.KEYWORD.value and isinstance(
        value,
        str,
    ):
        updated_value = process_keyword(value)
    elif metadata_identifier == DatasetIdentifiers.VERSION.value:
        updated_value = str(value)
    elif metadata_identifier in MULTIPLE_LANGUAGE_DATASET_METADATA and isinstance(
        value,
        str,
    ):
        updated_value = find_existing_language_string(
            state.metadata.dataset,
            value,
            metadata_identifier,
        )
    else:
        updated_value = value

    # Other values get returned unchanged
    return updated_value


def accept_dataset_metadata_input(
    value: MetadataInputTypes | LanguageStringType,
    metadata_identifier: str,
) -> tuple[bool, str]:
    """Handle user inputs of dataset metadata values."""
    logger.debug(
        "Received updated value = %s for metadata_identifier = %s",
        value,
        metadata_identifier,
    )
    try:
        value = process_special_cases(value, metadata_identifier)
        # Update the value in the model
        setattr(
            state.metadata.dataset,
            metadata_identifier,
            value,
        )
    except (ValidationError, ValueError) as e:
        show_error = True
        error_explanation = str(e)
        logger.exception("Error while reading in value for %s", metadata_identifier)
    else:
        show_error = False
        error_explanation = ""
        logger.debug(
            "Successfully updated value = %s for metadata_identifier = %s",
            value,
            metadata_identifier,
        )

    return show_error, error_explanation


def update_dataset_metadata_language() -> list[MetadataInputTypes]:
    """Return new values for ALL the dataset metadata inputs.

    This allows editing of strings in the chosen language.
    """
    return [
        m.value_getter(state.metadata.dataset, m.identifier)
        for m in DISPLAYED_DATASET_METADATA
    ]


def change_language_dataset_metadata(
    language: SupportedLanguages,
) -> tuple[object, ...]:
    """Change the language for the displayed dataset metadata.

    This is done in three steps:
    - Update the chosen language globally.
    - Update the language for dropdown options.
    - Update the language of all the metadata values.
    """
    update_global_language_state(language)
    return (
        *(e.options_getter(language) for e in DISPLAYED_DROPDOWN_DATASET_METADATA),
        update_dataset_metadata_language(),
    )


def accept_dataset_metadata_date_input(
    dataset_identifier: DatasetIdentifiers,
    contains_data_from: str | None,
    contains_data_until: str | None,
) -> tuple[bool, str, bool, str]:
    """Validate and save date range inputs."""
    try:
        (
            parsed_contains_data_from,
            parsed_contains_data_until,
        ) = parse_and_validate_dates(
            str(contains_data_from),
            str(contains_data_until),
        )

        if parsed_contains_data_from:
            state.metadata.dataset.contains_data_from = (
                parsed_contains_data_from.isoformat()
            )

        if parsed_contains_data_until:
            state.metadata.dataset.contains_data_until = (
                parsed_contains_data_until.isoformat()
            )

    except (ValidationError, ValueError) as e:
        logger.exception(
            "Validation failed for %s, %s, %s: %s, %s",
            dataset_identifier,
            "contains_data_from",
            contains_data_from,
            "contains_data_until",
            contains_data_until,
        )
        message: str | None = str(e)
    else:
        logger.debug(
            "Successfully updated %s, %s, %s: %s, %s",
            dataset_identifier,
            "contains_data_from",
            contains_data_from,
            "contains_data_until",
            contains_data_until,
        )
        message = None

    no_error = (False, "")
    if not message:
        # No error to display.
        return no_error + no_error

    error = (True, message)
    return (
        error + no_error
        if dataset_identifier == DatasetIdentifiers.CONTAINS_DATA_FROM
        else no_error + error
    )



"""Handle reading, updating and writing of metadata."""

from __future__ import annotations

import json
import logging
import pathlib
import uuid
from typing import TYPE_CHECKING

from cloudpathlib import CloudPath
from cloudpathlib import GSClient
from cloudpathlib import GSPath
from dapla import AuthClient
from datadoc_model import model

from datadoc.backend import user_info
from datadoc.backend.dapla_dataset_path_info import DaplaDatasetPathInfo
from datadoc.backend.dataset_parser import DatasetParser
from datadoc.backend.model_backwards_compatibility import upgrade_metadata
from datadoc.enums import Assessment
from datadoc.enums import DataSetState
from datadoc.enums import DataSetStatus
from datadoc.frontend.fields.display_dataset import (
    OBLIGATORY_DATASET_METADATA_IDENTIFIERS,
)
from datadoc.frontend.fields.display_variables import (
    OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS,
)
from datadoc.utils import METADATA_DOCUMENT_FILE_SUFFIX
from datadoc.utils import calculate_percentage
from datadoc.utils import get_timestamp_now

if TYPE_CHECKING:
    from datetime import datetime

    from datadoc.backend.statistic_subject_mapping import StatisticSubjectMapping

logger = logging.getLogger(__name__)


class DataDocMetadata:
    """Handle reading, updating and writing of metadata."""

    def __init__(
        self,
        statistic_subject_mapping: StatisticSubjectMapping,
        dataset_path: str | None = None,
        metadata_document_path: str | None = None,
    ) -> None:
        """Read in a dataset if supplied, otherwise naively instantiate the class."""
        self._statistic_subject_mapping = statistic_subject_mapping
        self.metadata_document: pathlib.Path | CloudPath | None = None
        self.container: model.MetadataContainer | None = None
        self.dataset_path: pathlib.Path | CloudPath | None = None
        self.short_name: str | None = None
        self.dataset = model.Dataset()
        self.variables: list = []
        self.variables_lookup: dict[str, model.Variable] = {}
        if metadata_document_path:
            # In this case the user has specified an independent metadata document for editing
            # without a dataset.
            self.metadata_document = self._open_path(metadata_document_path)
        elif dataset_path:
            self.dataset_path = self._open_path(dataset_path)
            # Build the metadata document path based on the dataset path
            # Example: /path/to/dataset.parquet -> /path/to/dataset__DOC.json
            self.metadata_document = self.dataset_path.parent / (
                self.dataset_path.stem + METADATA_DOCUMENT_FILE_SUFFIX
            )
        self.extract_metadata_from_files()

    @staticmethod
    def _open_path(path: str) -> pathlib.Path | CloudPath:
        """Open a given path regardless of whether it is local or cloud.

        The returned path may be treated just as if it's a pathlib.Path.
        """
        if path.startswith(GSPath.cloud_prefix):
            client = GSClient(credentials=AuthClient.fetch_google_credentials())
            return GSPath(path, client=client)

        return pathlib.Path(path)

    def extract_metadata_from_files(self) -> None:
        """Read metadata from an existing metadata document.

        If no metadata document exists, create one from scratch by extracting metadata
        from the dataset file.
        """
        if self.metadata_document is not None and self.metadata_document.exists():
            self.extract_metadata_from_existing_document(self.metadata_document)

        if (
            self.dataset_path is not None
            and self.dataset == model.Dataset()
            and len(self.variables) == 0
        ):
            self.extract_metadata_from_dataset(self.dataset_path)
            self.dataset.id = uuid.uuid4()
            # Set default values for variables where appropriate
            v: model.Variable
            for v in self.variables:
                if v.variable_role is None:
                    v.variable_role = model.VariableRole.MEASURE
                if v.direct_person_identifying is None:
                    v.direct_person_identifying = False
        if not self.dataset.id:
            self.dataset.id = uuid.uuid4()
        self.variables_lookup = {v.short_name: v for v in self.variables}

    def extract_metadata_from_existing_document(
        self,
        document: pathlib.Path | CloudPath,
    ) -> None:
        """There's an existing metadata document, so read in the metadata from that."""
        fresh_metadata = {}
        try:
            with document.open(mode="r", encoding="utf-8") as file:
                fresh_metadata = json.load(file)
            logger.info("Opened existing metadata file %s", document)
            if self.is_metadata_in_container_structure(fresh_metadata):
                self.container = model.MetadataContainer.model_validate_json(
                    json.dumps(fresh_metadata),
                )
                datadoc_metadata = fresh_metadata["datadoc"]
            else:
                datadoc_metadata = fresh_metadata
            if datadoc_metadata is None:
                # In this case we've read in a file with an empty "datadoc" structure.
                # A typical example of this is a file produced from a pseudonymization process.
                return
            datadoc_metadata = upgrade_metadata(
                datadoc_metadata,
            )
            meta = model.DatadocMetadata.model_validate_json(
                json.dumps(datadoc_metadata),
            )
            if meta.dataset is not None:
                self.dataset = meta.dataset
            if meta.variables is not None:
                self.variables = meta.variables
        except json.JSONDecodeError:
            logger.warning(
                "Could not open existing metadata file %s. \
                    Falling back to collecting data from the dataset",
                document,
                exc_info=True,
            )

    def is_metadata_in_container_structure(
        self,
        metadata: dict,
    ) -> bool:
        """At a certain point a metadata 'container' was introduced.

        The container provides a structure for different 'types' of metadata, such as 'datadoc', 'pseudonymization' etc.
        This method returns True if the metadata is in the container structure, False otherwise.
        """
        return "datadoc" in metadata

    def extract_metadata_from_dataset(
        self,
        dataset: pathlib.Path | CloudPath,
    ) -> None:
        """Obtain what metadata we can from the dataset itself.

        This makes it easier for the user by 'pre-filling' certain fields.
        Certain elements are dependent on the dataset being saved according to SSB's standard.
        """
        self.ds_schema: DatasetParser = DatasetParser.for_file(dataset)
        dapla_dataset_path_info = DaplaDatasetPathInfo(dataset)

        subject_field = self._statistic_subject_mapping.get_secondary_subject(
            dapla_dataset_path_info.statistic_short_name,
        )

        self.dataset = model.Dataset(
            short_name=self.dataset_path.stem if self.dataset_path else None,
            dataset_state=dapla_dataset_path_info.dataset_state,
            dataset_status=DataSetStatus.DRAFT,
            assessment=self.get_assessment_by_state(
                dapla_dataset_path_info.dataset_state,
            ),
            version=dapla_dataset_path_info.dataset_version,
            contains_data_from=str(dapla_dataset_path_info.contains_data_from),
            contains_data_until=str(dapla_dataset_path_info.contains_data_until),
            data_source_path=self.dataset_path,
            metadata_created_by=user_info.get_user_info_for_current_platform().short_email,
            # TODO @mmwinther: Remove multiple_language_support once the model is updated.
            # https://github.com/statisticsnorway/ssb-datadoc-model/issues/41
            subject_field=model.LanguageStringType(
                en=subject_field,
                nb=subject_field,
                nn=subject_field,
            ),
        )
        self.variables = self.ds_schema.get_fields()

    @staticmethod
    def get_assessment_by_state(state: DataSetState | None) -> Assessment | None:
        """Find assessment derived by dataset state."""
        if state is None:
            return None
        match (state):
            case (
                DataSetState.INPUT_DATA
                | DataSetState.PROCESSED_DATA
                | DataSetState.STATISTICS
            ):
                return Assessment.PROTECTED
            case DataSetState.OUTPUT_DATA:
                return Assessment.OPEN
            case _:
                return None

    def write_metadata_document(self) -> None:
        """Write all currently known metadata to file."""
        timestamp: datetime = get_timestamp_now()

        if self.dataset.metadata_created_date is None:
            self.dataset.metadata_created_date = timestamp
        self.dataset.metadata_last_updated_date = timestamp
        self.dataset.metadata_last_updated_by = (
            user_info.get_user_info_for_current_platform().short_email
        )
        self.dataset.file_path = str(self.dataset_path)

        datadoc: model.DatadocMetadata = model.DatadocMetadata(
            percentage_complete=self.percent_complete,
            dataset=self.dataset,
            variables=self.variables,
        )
        if self.container:
            self.container.datadoc = datadoc
        else:
            self.container = model.MetadataContainer(datadoc=datadoc)
        if self.metadata_document:
            self.metadata_document.write_text(self.container.model_dump_json(indent=4))
            logger.info("Saved metadata document %s", self.metadata_document)
        else:
            msg = "No metadata document to save"
            raise ValueError(msg)

    @property
    def percent_complete(self) -> int:
        """The percentage of obligatory metadata completed.

        A metadata field is counted as complete when any non-None value is
        assigned. Used for a live progress bar in the UI, as well as being
        saved in the datadoc as a simple quality indicator.
        """
        num_all_fields = len(OBLIGATORY_DATASET_METADATA_IDENTIFIERS)
        num_set_fields = len(
            [
                k
                for k, v in self.dataset.model_dump().items()
                if k in OBLIGATORY_DATASET_METADATA_IDENTIFIERS and v is not None
            ],
        )
        for variable in self.variables:
            num_all_fields += len(OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS)
            num_set_fields += len(
                [
                    k
                    for k, v in variable.model_dump().items()
                    if k in OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS and v is not None
                ],
            )
        return calculate_percentage(num_set_fields, num_all_fields)



"""Functions which aren't directly called from a decorated callback."""

from __future__ import annotations

import datetime
import logging
from typing import TYPE_CHECKING
from typing import TypeAlias
from typing import cast

import arrow
from datadoc_model import model

from datadoc import config
from datadoc import enums
from datadoc import state

if TYPE_CHECKING:
    import pathlib
    from enum import Enum

    import pydantic
    from cloudpathlib import CloudPath

    from datadoc.enums import SupportedLanguages

logger = logging.getLogger(__name__)


MetadataInputTypes: TypeAlias = (
    str | list[str] | int | float | bool | datetime.date | None
)


def update_global_language_state(language: SupportedLanguages) -> None:
    """Update global language state."""
    if state.current_metadata_language != language:
        logger.debug("Updating language: %s", language.name)
        state.current_metadata_language = language


def get_language_strings_enum(
    enum: Enum | type[enums.LanguageStringsEnum],
) -> enums.LanguageStringsEnum:
    """Get the correct language strings enum for the given enum.

    We need multiple languages to display in the front end, but the model only defines a single language in the enums.
    """
    language_strings_enum: enums.LanguageStringsEnum = getattr(
        enums,
        cast(type[enums.LanguageStringsEnum], enum).__name__,
    )
    return language_strings_enum


def get_options_for_language(
    language: SupportedLanguages,
    enum: Enum,
) -> list[dict[str, str]]:
    """Generate the list of options based on the currently chosen language."""
    return [
        {
            "label": i.get_value_for_language(language),
            "value": i.name,
        }
        for i in get_language_strings_enum(enum)  # type: ignore [arg-type, attr-defined]
    ]


def find_existing_language_string(
    metadata_model_object: pydantic.BaseModel,
    value: str,
    metadata_identifier: str,
) -> model.LanguageStringType:
    """Get or create a LanguageStrings object and return it."""
    # In this case we need to set the string to the correct language code
    language_strings = getattr(metadata_model_object, metadata_identifier)
    if language_strings is None:
        # This means that no strings have been saved yet so we need to construct
        # a new LanguageStrings object
        language_strings = model.LanguageStringType(
            **{state.current_metadata_language.value: value},
        )
    else:
        # In this case there's an existing object so we save this string
        # to the current language
        setattr(language_strings, state.current_metadata_language.value, value)
    return language_strings


def get_dataset_path() -> pathlib.Path | CloudPath | str | None:
    """Extract the path to the dataset from the potential sources."""
    if state.metadata.dataset_path is not None:
        return state.metadata.dataset_path
    path_from_env = config.get_datadoc_dataset_path()
    logger.info(
        "Dataset path from env var: '%s'",
        path_from_env,
    )
    return path_from_env


ISO_DATE_FORMAT = "YYYY-MM-DD"

VALIDATION_ERROR = "Validation error: "
DATE_VALIDATION_MESSAGE = f"{VALIDATION_ERROR}contains_data_from must be the same or earlier date than contains_data_until"


def parse_and_validate_dates(
    start_date: str | datetime.datetime | None,
    end_date: str | datetime.datetime | None,
) -> tuple[datetime.datetime | None, datetime.datetime | None]:
    """Parse and validate the given dates.

     Validate that:
         - The dates are in YYYY-MM-DD format
         - The start date is earlier or identical to the end date.

    Examples:
    >>> parse_and_validate_dates("2021-01-01", "2021-01-01")
    (datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc))

    >>> parse_and_validate_dates("1990-01-01", "2050-01-01")
    (datetime.datetime(1990, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), datetime.datetime(2050, 1, 1, 0, 0, tzinfo=datetime.timezone.utc))

    >>> parse_and_validate_dates(None, None)
    (None, None)

    >>> parse_and_validate_dates("1st January 2021", "1st January 2021")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: Expected an ISO 8601-like string, but was given '1st January 2021'. Try passing in a format string to resolve this.

    >>> parse_and_validate_dates(datetime.datetime(2050, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), "1990-01-01")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: contains_data_from must be the same or earlier date than contains_data_until

    >>> parse_and_validate_dates("2050-01-01", "1990-01-01")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: contains_data_from must be the same or earlier date than contains_data_until
    """
    parsed_start = None
    parsed_end = None
    try:
        if start_date and start_date != "None":
            parsed_start = arrow.get(start_date)
        if end_date and end_date != "None":
            parsed_end = arrow.get(end_date)
    except arrow.parser.ParserError as e:
        raise ValueError(VALIDATION_ERROR + str(e)) from e

    if parsed_start and parsed_end and (parsed_start > parsed_end):
        raise ValueError(DATE_VALIDATION_MESSAGE)

    start_output = (
        parsed_start.astimezone(tz=datetime.timezone.utc) if parsed_start else None
    )
    end_output = parsed_end.astimezone(tz=datetime.timezone.utc) if parsed_end else None

    return start_output, end_output



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from datadoc.backend.external_sources.external_sources import GetExternalSource
from datadoc.enums import SupportedLanguages

if TYPE_CHECKING:
    import concurrent

    import pandas as pd

from klass.classes.classification import KlassClassification

logger = logging.getLogger(__name__)


@dataclass
class CodeListItem:
    """Data structure for a code list item."""

    titles: dict[str, str]
    code: str

    def get_title(self, language: SupportedLanguages) -> str:
        """Get the title in the given language."""
        try:
            return self.titles[language]
        except KeyError:
            try:
                return self.titles[
                    (
                        "nb"
                        if language
                        in [
                            SupportedLanguages.NORSK_BOKMÅL,
                            SupportedLanguages.NORSK_NYNORSK,
                        ]
                        else "en"
                    )
                ]
            except KeyError:
                logger.exception(
                    "Could not find title for subject %s  and language: %s",
                    self,
                    language.name,
                )
                return ""


class CodeList(GetExternalSource):
    """Class for retrieving classifications from Klass."""

    def __init__(
        self,
        executor: concurrent.futures.ThreadPoolExecutor,
        classification_id: int | None,
    ) -> None:
        """Retrieves a list of classifications given a classification id.

        Initializes the classifications list and starts fetching the classifications.
        """
        self.supported_languages = [
            SupportedLanguages.NORSK_BOKMÅL.value,
            SupportedLanguages.ENGLISH.value,
        ]
        self._classifications: list[CodeListItem] = []
        self.classification_id = classification_id
        self.classifications_dataframes: dict[str, pd.DataFrame] | None = None
        super().__init__(executor)

    def _fetch_data_from_external_source(
        self,
    ) -> dict[str, pd.DataFrame] | None:
        """Fetches the classifications from Klass by classification id.

        returns a pandas dataframe with the class data for the given classification id.
        """
        try:
            classifications_dataframes = {}
            for i in self.supported_languages:
                classifications_dataframes[i] = (
                    KlassClassification(
                        str(self.classification_id),
                        i,
                    )
                    .get_codes()
                    .data
                )
        except Exception:
            logger.exception(
                "Exception while getting classifications from Klass",
            )
            return None
        else:
            return classifications_dataframes

    def _extract_titles(
        self,
        dataframes: dict[SupportedLanguages, pd.DataFrame],
    ) -> list[dict[str, str]]:
        list_of_titles = []
        languages = list(dataframes)
        for i in range(len(dataframes[SupportedLanguages.NORSK_BOKMÅL])):
            titles = {}
            for j in languages:
                if "name" in dataframes[j]:
                    titles[str(j)] = dataframes[j].loc[:, "name"][i]
                else:
                    titles[str(j)] = None
            list_of_titles.append(titles)
        return list_of_titles

    def _create_code_list_from_dataframe(
        self,
        classifications_dataframes: dict[SupportedLanguages, pd.DataFrame],
    ) -> list[CodeListItem]:
        """Method that finds the name column in the dataframe, and returns all values in a list."""
        classification_names = self._extract_titles(classifications_dataframes)
        classification_codes: list
        if "code" in classifications_dataframes[SupportedLanguages.NORSK_BOKMÅL]:
            classification_codes = (
                classifications_dataframes[SupportedLanguages.NORSK_BOKMÅL]
                .loc[:, "code"]
                .to_list()
            )
        else:
            classification_codes = [None] * len(classification_names)
        unit_types = []
        for a, b in zip(classification_names, classification_codes):
            unit_types.append(
                CodeListItem(a, b),
            )
        return unit_types

    def _get_classification_dataframe_if_loaded(self) -> bool:
        """Checks if the data from Klass is loaded, then gets the classifications from the dataframe."""
        if not self._classifications:
            self.classifications_dataframes = self.retrieve_external_data()
            if self.classifications_dataframes is not None:
                self._classifications = self._create_code_list_from_dataframe(
                    self.classifications_dataframes,
                )
                logger.debug(
                    "Thread finished. found %s classifications",
                    len(self._classifications),
                )
                return True
            logger.warning(
                "Thread is not done. Cannot get classifications from the dataframe.",
            )
        return False

    @property
    def classifications(self) -> list[CodeListItem]:
        """Getter for primary subjects."""
        self._get_classification_dataframe_if_loaded()
        logger.debug("Got %s classifications subjects", len(self._classifications))
        return self._classifications



from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeVar

if TYPE_CHECKING:
    import concurrent.futures

logger = logging.getLogger(__name__)

T = TypeVar("T")


class GetExternalSource(ABC, Generic[T]):
    """Abstract base class for getting data from external sources."""

    def __init__(self, executor: concurrent.futures.ThreadPoolExecutor) -> None:
        """Retrieves data from an external source asynchronously.

        Initializes the future object.
        """
        self.future = executor.submit(
            self._fetch_data_from_external_source,
        )

    def wait_for_external_result(self) -> None:
        """Waits for the thread responsible for loading the external request to finish."""
        if not self.future:
            logger.warning("No future to wait for.")
            # Nothing to wait for in this case, just return immediately
            return
        self.future.result()

    def check_if_external_data_is_loaded(self) -> bool:
        """Method to check if the thread getting the external data has finished running."""
        if self.future:
            return self.future.done()
        return False

    def retrieve_external_data(self) -> T | None:
        """Method that returns the result of the thread."""
        if self.future:
            return self.future.result()
        return None

    @abstractmethod
    def _fetch_data_from_external_source(self) -> T | None:
        """Abstract method implemented in the child class to handle the external data."""
        raise NotImplementedError



"""Top-level entrypoint, configuration and layout for the datadoc app.

Members of this module should not be imported into any sub-modules, this will cause circular imports.
"""

from __future__ import annotations

import concurrent
import logging
from pathlib import Path

import dash_bootstrap_components as dbc
from dash import Dash
from dash import html
from flask_healthz import healthz

from datadoc import config
from datadoc import state
from datadoc.backend.code_list import CodeList
from datadoc.backend.datadoc_metadata import DataDocMetadata
from datadoc.backend.statistic_subject_mapping import StatisticSubjectMapping
from datadoc.enums import SupportedLanguages
from datadoc.frontend.callbacks.register_callbacks import register_callbacks
from datadoc.frontend.components.alerts import dataset_validation_error
from datadoc.frontend.components.alerts import opened_dataset_error
from datadoc.frontend.components.alerts import opened_dataset_success
from datadoc.frontend.components.alerts import saved_metadata_success
from datadoc.frontend.components.alerts import variables_validation_error
from datadoc.frontend.components.control_bars import build_controls_bar
from datadoc.frontend.components.control_bars import build_language_dropdown
from datadoc.frontend.components.control_bars import header
from datadoc.frontend.components.control_bars import progress_bar
from datadoc.frontend.components.dataset_tab import build_dataset_tab
from datadoc.frontend.components.variables_tab import build_variables_tab
from datadoc.logging_configuration.logging_config import configure_logging
from datadoc.utils import get_app_version
from datadoc.utils import pick_random_port
from datadoc.utils import running_in_notebook

configure_logging()
logger = logging.getLogger(__name__)


def build_app(app: type[Dash]) -> Dash:
    """Define the layout, register callbacks."""
    tabs_children = [
        build_dataset_tab(),
        build_variables_tab(),
    ]

    app.layout = html.Div(
        children=[
            html.Header(
                [
                    header,
                ],
                className="header-wrapper",
            ),
            html.Main(
                [
                    progress_bar,
                    build_controls_bar(),
                    variables_validation_error,
                    dataset_validation_error,
                    opened_dataset_error,
                    saved_metadata_success,
                    opened_dataset_success,
                    dbc.Tabs(
                        id="tabs",
                        class_name="ssb-tabs",
                        children=tabs_children,
                    ),
                ],
                className="main-content-app",
            ),
            build_language_dropdown(),
        ],
        className="app-wrapper",
    )

    register_callbacks(app)

    return app


def get_app(
    executor: concurrent.futures.ThreadPoolExecutor,
    dataset_path: str | None = None,
) -> tuple[Dash, int]:
    """Centralize all the ugliness around initializing the app."""
    logger.info("Datadoc version v%s", get_app_version())
    collect_data_from_external_sources(executor)
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL
    state.metadata = DataDocMetadata(
        state.statistic_subject_mapping,
        dataset_path=dataset_path,
    )

    # The service prefix must be set to run correctly on Dapla Jupyter
    if prefix := config.get_jupyterhub_service_prefix():
        port = pick_random_port()
        requests_pathname_prefix = f"{prefix}proxy/{port}/"
    else:
        port = config.get_port()
        requests_pathname_prefix = "/"

    name = config.get_app_name()

    app = Dash(
        name=name,
        title=name,
        assets_folder=f"{Path(__file__).parent}/assets",
        requests_pathname_prefix=requests_pathname_prefix,
    )
    app = build_app(app)
    app.server.register_blueprint(healthz, url_prefix="/healthz")
    app.server.config["HEALTHZ"] = {
        "live": lambda: True,
        "ready": lambda: True,
        "startup": lambda: True,
    }
    logger.info("Built app with endpoints configured on /healthz")

    return app, port


def collect_data_from_external_sources(
    executor: concurrent.futures.ThreadPoolExecutor,
) -> None:
    """Call classes and methods which collect data from external sources.

    Must be non-blocking to prevent delays in app startup.
    """
    logger.debug("Start threads - Collecting data from external sources")
    state.statistic_subject_mapping = StatisticSubjectMapping(
        executor,
        config.get_statistical_subject_source_url(),
    )

    state.unit_types = CodeList(
        executor,
        config.get_unit_code(),
    )

    state.organisational_units = CodeList(
        executor,
        config.get_organisational_unit_code(),
    )
    logger.debug("Finished blocking - Collecting data from external sources")


def main(dataset_path: str | None = None) -> None:
    """Entrypoint when running as a script."""
    if dataset_path:
        logger.info("Starting app with dataset_path = %s", dataset_path)

    with concurrent.futures.ThreadPoolExecutor(max_workers=12) as executor:
        app, port = get_app(executor, dataset_path)
        if running_in_notebook():
            logger.info("Running in notebook")
            app.run(
                jupyter_mode="tab",
                jupyter_server_url=config.get_jupyterhub_http_referrer(),
                jupyter_height=1000,
                port=port,
            )
        else:
            if dev_mode := config.get_dash_development_mode():
                logger.warning(
                    "Starting in Development Mode. NOT SUITABLE FOR PRODUCTION.",
                )
            app.run(debug=dev_mode, port=port)


if __name__ == "__main__":
    main()



"""Components and layout for the Dataset metadata tab."""

from __future__ import annotations

from typing import TYPE_CHECKING

import ssb_dash_components as ssb
from dash import html

from datadoc.frontend.components.builders import build_ssb_styled_tab

if TYPE_CHECKING:
    import dash_bootstrap_components as dbc

SECTION_WRAPPER_ID = "section-wrapper-id"


def build_dataset_tab() -> dbc.Tab:
    """Build the Dataset metadata tab."""
    return build_ssb_styled_tab(
        "Datasett",
        html.Article(
            [
                html.Header(
                    [
                        ssb.Title(
                            "Datasett detaljer",
                            size=2,
                            className="workspace-title",
                        ),
                    ],
                    className="workspace-header",
                ),
                html.Article(
                    id=SECTION_WRAPPER_ID,
                    className="workspace-content",
                ),
            ],
            className="workspace-page-wrapper",
        ),
    )



"""Callback functions to do with variables metadata."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pydantic import ValidationError

from datadoc import state
from datadoc.frontend.callbacks.utils import MetadataInputTypes
from datadoc.frontend.callbacks.utils import find_existing_language_string
from datadoc.frontend.callbacks.utils import parse_and_validate_dates
from datadoc.frontend.fields.display_variables import (
    MULTIPLE_LANGUAGE_VARIABLES_METADATA,
)
from datadoc.frontend.fields.display_variables import VariableIdentifiers

if TYPE_CHECKING:
    from datadoc_model.model import LanguageStringType

logger = logging.getLogger(__name__)


def handle_multi_language_metadata(
    metadata_field: str,
    new_value: MetadataInputTypes | LanguageStringType,
    updated_row_id: str,
) -> MetadataInputTypes | LanguageStringType:
    """Handle updates to fields which support multiple languages."""
    if new_value is None:
        # This edge case occurs when the user removes the text in an input field
        # We want to ensure we only remove the content for the current language,
        # not create a new blank object!
        return find_existing_language_string(
            state.metadata.variables_lookup[updated_row_id],
            "",
            metadata_field,
        )

    if isinstance(new_value, str):
        return find_existing_language_string(
            state.metadata.variables_lookup[updated_row_id],
            new_value,
            metadata_field,
        )

    return new_value


def accept_variable_metadata_input(
    value: MetadataInputTypes,
    variable_short_name: str,
    metadata_field: str,
) -> str | None:
    """Validate and save the value when variable metadata is updated.

    Returns an error message if an exception was raised, otherwise returns None.
    """
    logger.debug(
        "Updating %s, %s with %s",
        metadata_field,
        variable_short_name,
        value,
    )
    try:
        if metadata_field in MULTIPLE_LANGUAGE_VARIABLES_METADATA:
            new_value = handle_multi_language_metadata(
                metadata_field,
                value,
                variable_short_name,
            )
        elif value == "":
            # Allow clearing non-multiple-language text fields
            new_value = None
        else:
            new_value = value

        # Write the value to the variables structure
        setattr(
            state.metadata.variables_lookup[variable_short_name],
            metadata_field,
            new_value,
        )
    except (ValidationError, ValueError) as e:
        logger.exception(
            "Validation failed for %s, %s, %s:",
            metadata_field,
            variable_short_name,
            value,
        )
        return str(e)
    else:
        if value == "":
            value = None
        logger.debug(
            "Successfully updated %s, %s with %s",
            metadata_field,
            variable_short_name,
            value,
        )
        return None


def accept_variable_metadata_date_input(
    variable_identifier: VariableIdentifiers,
    variable_short_name: str,
    contains_data_from: str,
    contains_data_until: str,
) -> tuple[bool, str, bool, str]:
    """Validate and save date range inputs."""
    message = accept_variable_metadata_input(
        (
            contains_data_from
            if variable_identifier == VariableIdentifiers.CONTAINS_DATA_FROM
            else contains_data_until
        ),
        variable_short_name,
        variable_identifier,
    )

    try:
        (
            parsed_contains_data_from,
            parsed_contains_data_until,
        ) = parse_and_validate_dates(
            str(
                contains_data_from
                or state.metadata.variables_lookup[
                    variable_short_name
                ].contains_data_from,
            ),
            str(
                contains_data_until
                or state.metadata.variables_lookup[
                    variable_short_name
                ].contains_data_until,
            ),
        )

        # Save both values to the model if they pass validation.
        state.metadata.variables_lookup[
            variable_short_name
        ].contains_data_from = parsed_contains_data_from
        state.metadata.variables_lookup[
            variable_short_name
        ].contains_data_until = parsed_contains_data_until
    except (ValidationError, ValueError) as e:
        logger.exception(
            "Validation failed for %s, %s, %s: %s, %s: %s",
            variable_identifier,
            variable_short_name,
            "contains_data_from",
            contains_data_from,
            "contains_data_until",
            contains_data_until,
        )
        message = str(e)
    else:
        logger.debug(
            "Successfully updated %s, %s, %s: %s, %s: %s",
            variable_identifier,
            variable_short_name,
            "contains_data_from",
            contains_data_from,
            "contains_data_until",
            contains_data_until,
        )
        message = None

    no_error = (False, "")
    if not message:
        # No error to display.
        return no_error + no_error

    error = (True, message)
    return (
        error + no_error
        if variable_identifier == VariableIdentifiers.CONTAINS_DATA_FROM
        else no_error + error
    )



from __future__ import annotations

import logging.config
from typing import Any

from datadoc.config import get_log_level


def configure_logging(config: dict[str, Any] | None = None) -> None:
    """Configure logging for the application."""
    if not config:
        config = {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "simple": {
                    "format": "%(asctime)s.%(msecs)03d %(levelname)s %(name)s: %(message)s",
                    "logger": "name",
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                },
                "json": {
                    "()": "datadoc.logging_configuration.json_formatter.DatadocJSONFormatter",
                    "fmt_keys": {
                        "level": "levelname",
                        "message": "message",
                        "timestamp": "timestamp",
                        "logger": "name",
                        "module": "module",
                        "function": "funcName",
                        "line": "lineno",
                        "thread_name": "threadName",
                    },
                },
            },
            "handlers": {
                "stdout": {
                    "class": "logging.StreamHandler",
                    "level": get_log_level(),
                    "formatter": "simple",
                    "stream": "ext://sys.stdout",
                },
            },
            "loggers": {
                "root": {
                    "level": get_log_level(),
                    "handlers": [
                        "stdout",
                    ],
                },
            },
        }

    logging.config.dictConfig(config)



"""Components and layout for variables tab."""

from __future__ import annotations

from typing import TYPE_CHECKING

import ssb_dash_components as ssb
from dash import html

from datadoc.frontend.components.builders import build_ssb_styled_tab

if TYPE_CHECKING:
    import dash_bootstrap_components as dbc

VARIABLES_INFORMATION_ID = "variables-information"
ACCORDION_WRAPPER_ID = "accordion-wrapper"


def build_variables_tab() -> dbc.Tab:
    """Build the framework for the variables tab."""
    return build_ssb_styled_tab(
        "Variabler",
        html.Article(
            [
                html.Header(
                    [
                        ssb.Title(
                            "Variabel detaljer",
                            size=2,
                            className="workspace-title",
                        ),
                        ssb.Paragraph(
                            id=VARIABLES_INFORMATION_ID,
                            className="workspace-info-paragraph",
                        ),
                        ssb.Input(
                            label="Søk i variabler",
                            searchField=True,
                            disabled=True,
                            placeholder="Kommer...",
                            id="search-variables",
                            n_submit=0,
                            value="",
                        ),
                    ],
                    className="workspace-header",
                ),
                html.Article(
                    id=ACCORDION_WRAPPER_ID,
                    className="workspace-content",
                ),
            ],
            className="workspace-page-wrapper",
        ),
    )
